import sys
import torch
import numpy as np
import struct
import math
import json
from convert_utils import *

def get_param(model, name):
    return model[name].numpy()
    
def convert_attn(out_prefix, ln_out_prefix, prefix, model, f, split_q):
    if split_q:
        w = get_param(model, prefix + ".q_proj.weight")
        w = np.transpose(w)
        write_libnc_param(f, w, out_prefix + "q_attn/w")
        
        w = get_param(model, prefix + ".q_proj.bias")
        write_libnc_param(f, w, out_prefix + "q_attn/b")

        w_k = get_param(model, prefix + ".k_proj.weight")
        w_v = get_param(model, prefix + ".v_proj.weight")
        w = np.concatenate((w_k, w_v), axis = 0)
        w = np.transpose(w)
        write_libnc_param(f, w, out_prefix + "kv_attn/w")
        
        w_k = get_param(model, prefix + ".k_proj.bias")
        w_v = get_param(model, prefix + ".v_proj.bias")
        w = np.concatenate((w_k, w_v), axis = 0)
        write_libnc_param(f, w, out_prefix + "kv_attn/b")
    else:
        w_q = get_param(model, prefix + ".q_proj.weight")
        w_k = get_param(model, prefix + ".k_proj.weight")
        w_v = get_param(model, prefix + ".v_proj.weight")
        w = np.concatenate((w_q, w_k, w_v), axis = 0)
        w = np.transpose(w)
        write_libnc_param(f, w, out_prefix + "c_attn/w")
        
        w_q = get_param(model, prefix + ".q_proj.bias")
        w_k = get_param(model, prefix + ".k_proj.bias")
        w_v = get_param(model, prefix + ".v_proj.bias")
        w = np.concatenate((w_q, w_k, w_v), axis = 0)
        write_libnc_param(f, w, out_prefix + "c_attn/b")

    write_libnc_param(f, get_param(model, prefix + "_layer_norm.weight"), ln_out_prefix + "g")
    write_libnc_param(f, get_param(model, prefix + "_layer_norm.bias"), ln_out_prefix + "b")
    w = get_param(model, prefix + ".out_proj.weight")
    w = np.transpose(w)
    write_libnc_param(f, w, out_prefix + "c_proj/w")
    write_libnc_param(f, get_param(model, prefix + ".out_proj.bias"), out_prefix + "c_proj/b")

def convert_params(in_filename, out_filename):
    dict = torch.load(in_filename, map_location=torch.device('cpu'))
    model = dict["model"]
    args = dict["args"]
    d_model = model["encoder.layers.0.fc1.weight"].shape[1]
    d_inner = model["encoder.layers.0.fc1.weight"].shape[0]
    max_target_positions = 1024

    if d_model == 1024:
        d_key = 64
    else:
        d_key = 128
        
    encoder_layers = 1
    while "encoder.layers.{:d}.fc1.weight".format(encoder_layers) in model:
        encoder_layers += 1

    decoder_layers = 1
    while "decoder.layers.{:d}.fc1.weight".format(decoder_layers) in model:
        decoder_layers += 1
    
    f = open(out_filename, "wb")

    if model["decoder.embed_tokens.weight"].shape[0] == 256206:
        model_type = "nllb200"
    else:
        model_type = "m2m100"
        
    write_libnc_param_header(f, {"type": model_type, "data_type": "f16", "n_layer": decoder_layers, "n_encoder_layer": encoder_layers, "d_model": d_model, "d_key": d_key, "d_inner": d_inner });

    print("{:30} {:20} {:10}".format("NAME", "SHAPE", "TYPE"))

    #####################
    # encoder

# same as decoder    
#    w = get_param(model, "encoder.embed_tokens.weight")
#    write_libnc_param(f, w, "enc/wte")
    
    for layer_num in range(encoder_layers):
        prefix = "encoder.layers." + str(layer_num) + "."
        out_prefix = "enc/h" + str(layer_num) + "/"

        convert_attn(out_prefix + "attn/", out_prefix + "ln_1/", prefix + "self_attn", model, f, False)
        
        write_libnc_param(f, get_param(model, prefix + "final_layer_norm.weight"), out_prefix + "ln_2/g")
        write_libnc_param(f, get_param(model, prefix + "final_layer_norm.bias"), out_prefix + "ln_2/b")
        w = get_param(model, prefix + "fc1.weight")
        w = np.transpose(w)
        write_libnc_param(f, w, out_prefix + "mlp/c_fc/w")
        write_libnc_param(f, get_param(model, prefix + "fc1.bias"), out_prefix + "mlp/c_fc/b")
        w = get_param(model, prefix + "fc2.weight")
        w = np.transpose(w)
        write_libnc_param(f, w, out_prefix + "mlp/c_proj/w")
        write_libnc_param(f, get_param(model, prefix + "fc2.bias"), out_prefix + "mlp/c_proj/b")
        
    write_libnc_param(f, get_param(model, "encoder.layer_norm.weight"), "enc/ln_f/g")
    write_libnc_param(f, get_param(model, "encoder.layer_norm.bias"), "enc/ln_f/b")
    
    #####################
    # decoder

    # sinusoidal embeddings
    # from fairseq/utils.py make_positions(), the first position is 2 assuming padding_idx = 1.
    w = get_embedding(max_target_positions + 2, d_model)
    w = w[2:,:]
    write_libnc_param(f, w.numpy(), "wpe")

    w = get_param(model, "decoder.embed_tokens.weight")
    write_libnc_param(f, w, "wte")
    
    for layer_num in range(decoder_layers):
        prefix = "decoder.layers." + str(layer_num) + "."
        out_prefix = "h" + str(layer_num) + "/"

        convert_attn(out_prefix + "attn/", out_prefix + "ln_1/", prefix + "self_attn", model, f, False)
        
        convert_attn(out_prefix + "cross_attn/", out_prefix + "ln_0/", prefix + "encoder_attn", model, f, True)

        write_libnc_param(f, get_param(model, prefix + "final_layer_norm.weight"), out_prefix + "ln_2/g")
        write_libnc_param(f, get_param(model, prefix + "final_layer_norm.bias"), out_prefix + "ln_2/b")
        w = get_param(model, prefix + "fc1.weight")
        w = np.transpose(w)
        write_libnc_param(f, w, out_prefix + "mlp/c_fc/w")
        write_libnc_param(f, get_param(model, prefix + "fc1.bias"), out_prefix + "mlp/c_fc/b")
        w = get_param(model, prefix + "fc2.weight")
        w = np.transpose(w)
        write_libnc_param(f, w, out_prefix + "mlp/c_proj/w")
        write_libnc_param(f, get_param(model, prefix + "fc2.bias"), out_prefix + "mlp/c_proj/b")
        
    write_libnc_param(f, get_param(model, "decoder.layer_norm.weight"), "ln_f/g")
    write_libnc_param(f, get_param(model, "decoder.layer_norm.bias"), "ln_f/b")
    
    f.close()

if len(sys.argv) != 3:
    print("""usage: convert.py model_file output_file

Convert a M2M100 parameter dump to a LibNC one.

model_dir is the filename containing the PyTorch parameter dump
output_file is filename of the libNC parameter dump""")
    sys.exit(1)

convert_params(sys.argv[1], sys.argv[2])
